"""This is designed to let you do asynchronous I/O conveniently."""

from __future__ import with_statement

import threading

Lock = threading.Lock

class threading_with_result(threading.Thread):
	def __init__(self, tag=None, group=None, target=None, name=None, args=(), kwargs={}):
		threading.Thread.__init__(self, group=group, target=target, name=name, args=args, kwargs=kwargs)
		self.rv = None
		self.rvset = False
		self.tag = tag

	def run(self):
		self.rv = self._Thread__target(*self._Thread__args, **self._Thread__kwargs)
		self.rvset = True

	def get(self):
		assert self.rvset
		return self.rv


class Thread_poolW(object):
	"""Here, you have a pool of n threads.   You can call
	C{x.start(function, args, kwargs)} to start something running
	(perhaps a write to a file), and then go off and compute
	something else while the I/O completes.
	If you need to wait for completion, call C{x.join()}.

	This class is not designed to return values from the function.
	"""

	def __init__(self, n=4):
		self.n = n
		self.t = []


	def start(self, fn, args=None, kwargs=None):
		if len(self.t) >= self.n:
			for (i,t) in enumerate(self.t):
				if not t.isAlive():
					t.join()
					del self.t[i]
					break
		if len(self.t) >= self.n:
			self.t.pop(0).join()
		t = threading.Thread(target=fn, args=args, kwargs=kwargs)
		t.start()
		self.t.append(t)


	def join(self):
		for t in self.t:
			t.join()


Thread_poolNR = Thread_poolW	# Obsolete


class Thread_poolR(object):
	"""Here, you have a pool of n threads.   You can call
	C{x.start(function, args, kwargs)} to start something running
	(perhaps a write to a file), and then go off and compute
	something else while the I/O completes.
	If you need to wait for completion, call C{x.join()}.

	This class is not designed to return values from the function.
	"""

	def __init__(self, n=4):
		self.n = n
		self.t = []
		self.lock = threading.Lock()
		self.answers = []


	def start(self, tag, fn, args=None, kwargs=None):
		with self.lock:
			if len(self.t) >= self.n:
				self.__join_one()
			t = threading_with_result(tag=tag, target=fn, args=args, kwargs=kwargs)
			t.start()
			self.t.append(t)


	def __join_one(self):
		i = 0
		while i < len(self.t):
			t = self.t[i]
			if not t.isAlive():
				t.join()
				self.answers.append( (t.tag,t.get()) )
				del self.t[i]
			else:
				i += 1
		if self.t and not self.answers:
			t = self.t.pop(0)
			t.join()
			self.answers.append( (t.tag,t.get()) )




	def getany(self):
		with self.lock:
			self.__join_one()
			return self.answers.pop(0)


	def get(self, tag):
		with self.lock:
			for (i, (tg,v)) in enumerate(self.answers):
				if tg == tag:
					del self.answers[i]
					return v
			for (i,t) in enumerate(self.t):
				if t.tag == tag:
					t.join()
					del self.t[i]
					return t.get()
			raise KeyError, "Tag %s has not yet been started." % tag


def pairmap(fn, inlist, **kwargs):
	tp = Thread_poolR(kwargs.get('poolsize', 5))
	no = 0
	s = 0
	for (j,xi) in enumerate(inlist):
		tp.start(xi, fn, (xi,), kwargs)
		s += 1
		if s >= no+tp.n:
			yield tp.getany()
			no += 1
	while no < s:
		yield tp.getany()
		no += 1
	
	

def testmap():
	q = list(pairmap(lambda x:x, range(10)))
	assert len(q)==10
	seen = set()
	for (i,j) in q:
		assert i==j
		seen.add(i)
	assert sorted(seen)==range(10)


if __name__ == '__main__':
	testmap()
